"""
Pretty print a complete puppet documentation.

An ``output.json``, as per produced by ``./merge-json.py`` should be
provided as the first element. This program goes through every
definition in it, and outputs a complete index.html.
"""

from .markdown import markdown
import html
import re
from typing import (
    Tuple,
)

from .puppet.parser import puppet_parser
import logging

from .puppet.strings import (
    DataTypeAlias,
    DefinedType,
    DocString,
    Function,
    PuppetClass,
    ResourceType,
    DocStringParamTag,
    DocStringExampleTag,
)
from muppet.puppet.ast import build_ast
# from muppet.puppet.format import to_string
from muppet.parser_combinator import ParserCombinator, ParseError
from muppet.puppet.format.parser import ParserFormatter


logger = logging.getLogger(__name__)

# parse_puppet = puppet_parser

param_doc: dict[str, str] = {}


def parse_puppet(source: str, file: str) -> str:
    """
    Parse and syntax highlight the given puppet source.

    :returns: An HTML string
    """
    # logger.debug("source: %a", source)
    # Run the upstream puppet parser,
    # then masage the tree into a usable form.
    ast = build_ast(puppet_parser(source))
    # logger.info("ast: %a", ast)
    # From the ast, build a parser combinator parser.
    parser = ParserFormatter(source, file=file).serialize(ast)
    # logger.warning("parser: %a", parser)
    # Run the generatefd parser, giving us a list of match objects
    match_objects = ParserCombinator(source, file).get(parser)
    # logger.error("match_objects: %a", match_objects)
    return '\n'.join(repr(m) for m in match_objects)

# --------------------------------------------------


def format_docstring(name: str, docstring: DocString) -> Tuple[str, str]:
    """
    Format docstrings as they appear in some puppet types.

    Those types being:

    * puppet_classes,
    * puppet_type_aliases, and
    * defined_types
    """
    global param_doc

    out = ''

    param_doc = {tag.name: tag.text or ''
                 for tag in docstring.tags
                 if isinstance(tag, DocStringParamTag)}
    tags = docstring.tags

    # print(param_doc, file=sys.stderr)

    # param_defaults = d_type['defaults']

    for t in tags:
        text = html.escape(t.text)
        if t.tag_name == 'summary':
            out += '<em class="summary">'
            out += text
            out += '</em>'

    for t in tags:
        text = html.escape(t.text)
        if isinstance(t, DocStringExampleTag):
            if name := t.name:
                out += f'<h3>{name}</h3>\n'
            # TODO highlight?
            out += f'<pre class="example"><code class="puppet">{text}</code></pre>\n'

    # out += '<dl>'
    # for t in tags:
    #     if t['tag_name'] == 'param':
    #         out += f"<dt>{t['name']}</dt>"
    #         if text := t.get('text'):
    #             text = re.sub(r'(NOTE|TODO)',
    #                           r'<mark>\1</mark>',
    #                           markdown(text))
    #             out += f"<dd>{text}</dd>"
    # out += '</dl>'

    out += '<div>'
    out += markdown(docstring.text)
    out += '</div>'

    return (name, out)


# TODO @option tags
def build_param_dict(docstring: DocString) -> dict[str, str]:
    """
    Extract all parameter documentation from a docstring dict.

    :param docstring:
        The object present under 'docstring' in the information about
        a single object (class, resource, ...) in the output of
        `puppet strings`.

    :returns:
        A dictionary where the keys are the variables which have
        documentation, and the value is the (formatted) documentation
        for that key. Undocumented keys (even those with the tag, but
        no text) are ommitted from the resulting dictionary.
    """
    obj = {}
    for t in docstring.tags:
        if isinstance(t, DocStringParamTag):
            obj[t.name] = re.sub(r'(NOTE|TODO)',
                                 r'<mark>\1</mark>',
                                 markdown(t.text))
    return obj


def format_class(d_type: DefinedType | PuppetClass) -> Tuple[str, str]:
    """Format Puppet class."""
    out = ''
    logger.info("Formatting class %s", d_type.name)
    # print(name, file=sys.stderr)
    name, body = format_docstring(d_type.name, d_type.docstring)
    out += body

    out += '<pre class="highlight-muppet"><code class="puppet">'
    # ------ Old ---------------------------------------
    # t = parse_puppet(d_type.source)
    # data = parse(t, 0, ['root'])
    # renderer = HTMLRenderer(build_param_dict(d_type.docstring))
    # out += render(renderer, data)
    # ------ New ---------------------------------------
    try:
        out += parse_puppet(d_type.source, d_type.name)
    except ParseError as e:
        logger.error("Parsing %(name)s failed: %(err)s",
                     {'name': d_type.name, 'err': e})
    out += '</code></pre>'
    return name, out


def format_type() -> str:
    """Format Puppet type."""
    return 'TODO format_type not implemented'


def format_type_alias(d_type: DataTypeAlias, file: str) -> Tuple[str, str]:
    """Format Puppet type alias."""
    out = ''
    name = d_type.name
    logger.info("Formatting type alias %s", name)
    # print(name, file=sys.stderr)
    title, body = format_docstring(name, d_type.docstring)
    out += body
    out += '\n'
    out += '<pre class="highlight-muppet"><code class="puppet">'
    try:
        out += parse_puppet(d_type.alias_of, file)
    except ParseError as e:
        logger.error("Parsing %(name)s failed: %(err)s",
                     {'name': d_type.alias_of, 'err': e})
    out += '</code></pre>\n'
    return title, out


def format_defined_type(d_type: DefinedType, file: str) -> Tuple[str, str]:
    """Format Puppet defined type."""
    # renderer = HTMLRenderer(build_param_dict(d_type.docstring))
    out = ''
    name = d_type.name
    logger.info("Formatting defined type %s", name)
    # print(name, file=sys.stderr)
    title, body = format_docstring(name, d_type.docstring)
    out += body

    out += '<pre class="highlight-muppet"><code class="puppet">'
    try:
        out += parse_puppet(d_type.source, file)
    except ParseError as e:
        logger.error("Parsing %(name)s failed: %(err)s",
                     {'name': d_type.source, 'err': e})
    out += '</code></pre>\n'
    return title, out


def format_resource_type(r_type: ResourceType) -> str:
    """Format Puppet resource type."""
    name = r_type.name
    logger.info("Formatting resource type %s", name)
    out = ''
    out += f'<h2>{name}</h2>\n'
    out += str(r_type.docstring)

    out += '<h3>Properties</h3>\n'
    if props := r_type.properties:
        out += '<ul>\n'
        for property in props:
            out += f'<li>{property.name}</li>\n'
            # description, values, default
        out += '</ul>\n'
    else:
        out += '<em>No providers</em>'

    out += '<h3>Parameters</h3>\n'
    out += '<ul>\n'
    for parameter in r_type.parameters:
        out += f'<li>{parameter.name}</li>\n'
        # description
        # Optional[isnamevar]
    out += '</ul>\n'

    out += '<h3>Providers</h3>\n'
    if providers := r_type.providers:
        for provider in providers:
            out += f'<h4>{provider.name}</h4>\n'
            # TODO
    else:
        print('<em>No providers</em>')

    return out


def format_puppet_function(function: Function, file: str) -> str:
    """Format Puppet function."""
    out = ''
    name = function.name
    logger.info("Formatting puppet function %s", name)
    out += f'<h2>{name}</h2>\n'
    t = function.type
    # docstring = function.docstring
    for signature in function.signatures:
        signature.signature
        signature.docstring
    if t in ['ruby3x', 'ruby4x']:
        # TODO syntax highlighting
        s = '<pre class="highlight-muppet"><code class="ruby">'
        s += function.source
        s += '</code></pre>\n'
        out += s
    elif t == 'puppet':
        out += '<pre class="highlight-muppet"><code class="puppet">'
        try:
            out += parse_puppet(function.source, file)
        except ParseError as e:
            logger.error("Parsing %(name)s failed: %(err)s",
                         {'name': function.source, 'err': e})

        out += '</code></pre>\n'
    else:
        # TODO do something
        pass

    return out


def format_puppet_task() -> str:
    """Format Puppet task."""
    return 'TODO format_puppet_task not implemented'


def format_puppet_plan() -> str:
    """Format Puppet plan."""
    return 'TODO format_puppet_plan not implemented'
