"""
YAPC - Yet Another Parser Combinator.

This module implements a simple, and probably rather bad parser
combinator.

At its core is the ``ParserCombinator`` class, which is the driver for
the whole process. A sample usage is as follows:

.. code-block:: python

    filename = 'example.ext'
    with open(filename, 'r') as f:
        data = f.read()

    parser = ParserCombinator(data, file=filename)

    parser.get(grammar)

Where ``grammar`` is a parser  parser combinator rule.

All rules stem from the compound type ``Items`` (TODO which should be
renamed). The types of valid rules are:

- string literals, which match themselves
- ``None``, which matches nothing, and return nothing
- ParseDirective instances, see below
- Nullary functions, returing parser instances. This is purely for
  lazy evaluation.

    .. _sequence:

- Sequences of the above, interpreted as all the items after each
  other. Note that if earlier elements succeed, but later fails, then
  an exception will still be thrown, and the earlier parts discarded.
  See also: :ref:`and <and>`

``ParseDirective`` objects are the extension mechanism for this parser.

Rules
=====

Each parser MUST implement a ``run`` method, which should take an
instance of the parser object, and attempt to read characters from it.
Characters can be read from the parser instance through the ``read``
method. A couple of other usefull methods are available, see
:py:class:`ParserCombinator`.

A parser should either return a list of ``MatchObject``:s of what it
parsed, or throw a ``ParseError`` exception if the current parser
isn't applicable. These errors SHOULD be catched, since they should be
treated as non fatal. For exmaple

.. code-block:: python

    literal = integer | string

will be parsed by first trynig the integer parser, which SHOULD fail
if the input isn't a valid integer, and then try a string literal. If
neither matches, then the whole (``literal``) expression SHOULD throw.

Built in parsers
================

However, usually when constructing parser you instead want to build
from the "primitives" provided by this module. These include:

``optional(P)``
    Either matches the given parser, or matches nothing.

``count(P, n)``
    Matches exactly *n* instances of the parser *P*.

``discard(P)``
    Matches the parser *P*, but returns no matches.

``char``
    Matches any single character. This is the most basic parser possible.

``nop``
    Matches nothing. This is only useful as part of more complex parser.

``many(P)``
    Consume tokens until the given parser doesn't match any more, this
    means 0 or more.

``many1``
    Like ``many``, but at least one. Equivalent to ``P & many(P)``

``complement(P)``
    Parses anything except the given parser. Mainly useful together
    with ``all_``.

``delimited(P1, P2)``
    Parses a list of *P2*:s, delimited by the parser *P1*. A good
    usage could be:

    .. code-block:: python

        delimited(ws & "," & ws, integer)

    which would match a list of integers, separated by commas (and
    surrounding space).

``digit``
    Matches any one digit.

``hexdig``
    Matches any one hexadecimal digit.

``space``
    Matches a single ASCII whitespace character.

``ws``
    Matches any sequence of ``space``.

``all_(P ...)``
    Starting from the same point, run each parser in order. If all of
    them succeed, return the result of the last ran parser.

Meta commands
-------------

Besides these, a few "meta" commands exists, these include:

``s(any)``
    Which wraps its argument in a parser combinator object. Primarily
    useful for the operators (see below).

``name(name, P)``
    Attaches a name to the given parser. This causes the ``repr`` of
    the given object to the the given name.

``tag(tag, P)``
    Attaches the given tag to the parsed object. This is usefull to
    quickly figure out what kind of object was parsed. The following
    example is a good idea of how to implement an integer parser.

    .. code-block:: python

        integer = tag("integer", many1(number))

    When combining ``tag`` and ``name``, it's recommended to have
    ``name`` as the outermost, since otherwise the tag's ``repr``
    would still be visible.


Operators
---------

Finally, the base parser combinator class defines a few operators
(which mostly also are available as methods). These are:

.. _and:
``P1 & P2`` or ``and_(P1, P2)``
    Creates a new combinator, which matches if *P1* first matches,
    and *P2* matches wherever *P1* ends.

    Note that this is an all or nothing deal, in contrast to sequences
    from lists, which may fail partway through.

    See also: :ref:`sequence <sequence>`

``P1 | P2`` or ``or_(P1, P2)``)
    Creates a new combinator, which matches either *P1* or *P2*,
    returning the relut from the one matching.

``P @ f``
    Matches (or fails) *P* as usuall. However, if it matches, then the
    result will be passed through the procedure *f*, and its result
    will become part of the parsed tree instead of the initial result
    of *P*.

``~ P`` or ``not_(P)``
    Creates a combinator which matches the complement of *P*, meaning
    that this parser will be successfull if *P* fails.
"""

# import html
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Optional,
    Sequence,
    TypeAlias,
    Union,
    TypeVar,
)
import logging


logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class MatchCompound:
    """A matched item, similar to a regex match."""

    type: str
    matched: Any

    def __repr__(self) -> str:
        if self.type:
            return f'`{self.type}({self.matched!r})`'
        else:
            return f'MatchCompound({self.matched!r}))'


MatchObject: TypeAlias = MatchCompound | str


@dataclass
class ParseError(Exception):
    """
    Errors encountered while parsing.

    This should only appear with optional fields, since we don't know
    if the next token is the "expected" one. It should be captured
    internally by all exported procedures and methods.
    """

    msg: Optional[str] = None

    stk: list['Items'] = field(default_factory=list)

    def __str__(self) -> str:
        s = f"{self.msg}\nTraceback of called parsers:\n"
        for item in self.stk:
            s += '• '
            item = str(item)
            if len(item) > 60:
                s += item[:60] + '…'
            else:
                s += item
            s += '\n'
        return s


class ParseDirective:
    """
    Common type for special parsing directives.

    This is used for optional parsers, alternative parsers, and the like.
    """

    handler: Optional[Callable[[list[Any]], list[Any]]] = None

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:
        """
        Execute this directive.

        :param parser:
            The formatter which holds the context to use.
        """
        raise NotImplementedError(f'Missing run() for {self.__class__.__name__}')

    def __and__(self, other: Any) -> 'ParseDirective':
        return and_(self, other)

    def __or__(self, other: Any) -> 'ParseDirective':
        return or_(self, other)

    def __matmul__(self, proc: Any) -> 'ParseDirective':
        self.handler = proc
        return self

    def __invert__(self) -> 'ParseDirective':
        return not_(self)


# TODO rename this
Items: TypeAlias = Union[
    str,
    None,
    ParseDirective,
    Callable[[], 'Items'],
    Sequence['Items']
]


@dataclass
class s(ParseDirective):
    """
    Wraps a thingy in a parser directive.

    This directive exactly matches the given string, and is mostly
    here to wrap strings to allow infix operators.
    """

    s: Items

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        return parser.get(self.s)

    def __repr__(self) -> str:
        return repr(self.s)


@dataclass
class not_(ParseDirective):
    """Succeeds if the given directive fails."""

    form: Items

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        snapshot = parser.snapshot()
        try:
            # if this succeeds, then we want to fail
            # but if it fails then we want to suceed
            parser.get(self.form)
        except ParseError:
            parser.restore(snapshot)
            # The parser failed, meaning that we "suceeded"
            return []
        raise ParseError()

    def __repr__(self) -> str:
        return f'~ {self.form}'


@dataclass
class name(ParseDirective):
    """Attach a name to a parser, purely for debugging."""

    name: str
    form: 'Items'

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        return parser.get(self.form)

    def __repr__(self) -> str:
        return f'{self.name}'


def optional(parser: ParseDirective) -> ParseDirective:
    """Optionally parse parameter."""
    return parser | nop


@dataclass
class and_(ParseDirective):
    """Parse a group in order."""

    items: list['Items']

    def __init__(self, *items: 'Items'):
        self.items = list(items)

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        snapshot = parser.snapshot()
        out = []
        try:
            # print("self.items", self.items)
            for item in self.items:
                # print("item=", item)
                result = parser.get(item)
                # print("result=", result)
                out += result
        except ParseError as e:
            parser.restore(snapshot)
            raise e
        return out

    def __and__(self, other: Any) -> ParseDirective:
        if isinstance(other, and_):
            return and_(*self.items, *other.items)
        else:
            return and_(*self.items, other)

    def __repr__(self) -> str:
        return f'({" & ".join(repr(x) for x in self.items)})'


@dataclass
class or_(ParseDirective):
    """
    A set of multiple parse forms to try in order.

    The first matched one will be used.

    (Alias Variant)
    """

    alternatives: list['Items']

    def __init__(self, *alternatives: Items):
        """Retrun a new OptionParse object."""
        self.alternatives = list(alternatives)

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        save = parser.snapshot()
        for alternative in self.alternatives:
            try:
                return parser.get(alternative)
            except ParseError:
                parser.restore(save)
        else:
            msg = f"No alternative suceeded, cases={self.alternatives}, seek={parser.seek}"
            raise ParseError(msg)

    def __or__(self, other: Any) -> ParseDirective:
        if isinstance(other, or_):
            return or_(*self.alternatives, *other.alternatives)
        else:
            return or_(*self.alternatives, other)

    def __repr__(self) -> str:
        # return f'or_({", ".join(repr(x) for x in self.alternatives)})'
        return f'({" | ".join(repr(x) for x in self.alternatives)})'


@dataclass
class count(ParseDirective):
    """Run the given parser between min and max times."""

    min: int
    max: int
    parser: Items

    def __init__(self,
                 parser: Items,
                 arg: int,
                 *args: int,
                 **kwargs: int):
        """
        Run parser between ``min`` and ``max`` times.

        If only one numeric argument is given, then that's used as the
        maximum count, but if two values are given, then the minimum count
        preceeds the maximum.

        Both max and min can also be given as keyword arguments.

        :param parser:
        :param min:
        :param max:

        .. code-block:: python

            count(p, min, max)

        .. code-block:: python

            count(p, max)
        """
        min_ = 0
        max_ = 1

        match args:
            case []:
                max_ = arg
            case [v]:
                min_ = arg
                max_ = v

        if 'min' in kwargs:
            min_ = kwargs['min']
        if 'max' in kwargs:
            max_ = kwargs['max']

        self.min = min_
        self.max = max_
        self.parser = parser

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        snapshot = parser.snapshot()
        out = []
        try:
            # If any of these fails, then the whole operation was a fail.
            for _ in range(self.min):
                out += parser.get(self.parser)

            # These are optional, so an error here is "expected"
            for _ in range(self.min, self.max):
                try:
                    out += parser.get(self.parser)
                except ParseError:
                    break
        except Exception as e:
            out = []
            parser.restore(snapshot)
            raise e
        return out

    def __repr__(self) -> str:
        return f'count({repr(self.parser)}, {self.min}, {self.max})'


def discard(parser: Items) -> ParseDirective:
    """Run parser, but discard the result."""
    return s(parser) @ (lambda _: [])


@dataclass
class CharParser(ParseDirective):
    """Parse a single character."""

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        return [parser._get_char()]

    def __repr__(self) -> str:
        return 'char'


char = CharParser()


@dataclass
class all_(ParseDirective):
    """
    Run each parser in succession from the same point, returning the final result.

    .. code-block:: python

        all_(~space, char)
    """

    parsers: list[Items]

    def __init__(self, *parsers: Items) -> None:
        self.parsers = list(parsers)

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        match self.parsers:
            case [*ps, last]:
                for p in ps:
                    snapshot = parser.snapshot()
                    parser.get(p)
                    parser.restore(snapshot)
                return parser.get(last)
            case []:
                return []
        raise ValueError(f"This shoudn't be reachable: {self!r}")


@dataclass
class NopParser(ParseDirective):
    """Parses nothing."""

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        return []

    def __repr__(self) -> str:
        return 'nop'


nop = NopParser()


# @dataclass
# class many(ParseDirective):
#     """Many a parser as many times as possible."""
#
#     parser: 'Items'
#
#     def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
#         out = []
#         try:
#             while True:
#                 entry = parser.get(self.parser)
#                 if not entry:
#                     raise ParseError()
#                 # logger.info("seek = %s, parser = %s",
#                 #             parser.seek, self.parser)
#                 out += entry
#         except ParseError:
#             return out
#
#     def __repr__(self) -> str:
#         return f'many({repr(self.parser)})'


# This is a "nicer" looking implementation of ``many``, but it will
# probably blow up the call stack.


def many(parser: Items) -> ParseDirective:
    """Match a parser as many times as possible."""
    return (s(parser) & name('<rest>', lambda: many(parser))) @ (lambda xs: [xs[0], *xs[1:]]) \
        | nop


def many1(parser: ParseDirective) -> ParseDirective:
    """Parse between 1 any many."""
    return parser & many(parser)


@dataclass
class complement(ParseDirective):
    """Parses any character which isn't given."""

    chars: str

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        match parser.peek(char):
            case [str(c)]:
                if c not in self.chars:
                    return parser.get(char)
                else:
                    raise ParseError(msg=f"{c} in {self.chars}")
            case it:
                raise ParseError(msg=f"Parsed item wasn't a char: {it!r}")

    def __repr__(self) -> str:
        return f'complement({repr(self.chars)})'


@dataclass
class tag(ParseDirective):
    """Tag value returned by parser."""

    tag: str
    parser: 'Items'

    def run(self, parser: 'ParserCombinator') -> list[MatchObject]:  # noqa: D102
        result = parser.get(self.parser)
        return [MatchCompound(type=self.tag, matched=result)]


def delimited(delim: Items, parser: Items) -> ParseDirective:
    """
    Read an infix delimited list of items.

    If a optional trailing "comma" is wanted:

    .. code-block:: python

        and_(delimited(and_(ws, ',', ws),
                       ITEM),
            optional(ws, ','))
    """
    return s(parser) & many(s(delim) & s(parser))


digit = name('digit', or_(*(chr(x + ord('0')) for x in range(0, 10))))

hexdig = name('hexdig',
              digit
              | or_(*(chr(x + ord('A')) for x in range(0, 6)))
              | or_(*(chr(x + ord('a')) for x in range(0, 6))))

space = s(' ') | '\t' | '\n' | '\r'

ws = name('ws', many(space))


def line_comment(start: str) -> ParseDirective:
    """Match a line comment, from start delim to end of line."""
    return tag('line-comment', and_(start, many(complement('\n')), '\n'))


class ParserCombinatorSnapshot:
    """
    An opaque object storing the state of the ParserCombinator.

    This allows undoing any number of operations. Each parser snapshot
    can only be used with the parser formatter which created it.

    A snapshot can be restored multiple times.
    """

    def __init__(self, seek: int, creator: 'ParserCombinator'):
        self._seek = seek
        self._creator = creator


class ParserCombinator:
    """
    A basic parser combinator for Python.

    :param source:
        The string which should be parsed.
    :param file:
        Optional name of the file being parsed. Only used for debug purposes.
    """

    def __init__(self, source: str, file: Optional[str] = None):
        self.__source = source
        self.seek = 0
        self.file = file

    def peek(self, item: Items) -> list[MatchObject]:
        """Run the parser without updating the state."""
        # TODO on error?
        snapshot = self.snapshot()
        result = self.get(item)
        self.restore(snapshot)
        return result

    def _get_char(self) -> str:
        """
        Primitive for reading one character.

        This method reads one character from the buffer and returns
        it, or raises an IndexError if we are at the end of the
        buffer.

        This procedure shouldn't be called directly, but instead the
        parser ``char`` should be invoked.
        """
        try:
            out = self.__source[self.seek]
        except IndexError:
            raise ParseError("End of string")
        self.seek += 1
        return out

    def remaining(self) -> str:
        """Return remaining, unparsed, string."""
        return self.__source[self.seek:]

    def get(self, item: Items) -> list[Any]:
        """
        Try parsing the next item in stream as the passed on parser.

        :param item:
            Parser to run next.
        :returns:
            If the parsing suceeded, the list of matched objects are
            returned, and the interal stream is updated.

        :throws ParseError:
            If parsing failed, then a ParseError is thrown. However,
            if any tokens had been parsed that far then the internal
            stream is still upadted, and the matched objects are lost.
        """
        # logger.debug("item: %a", item)
        out: list[Any] = []
        # Step 1: decode our parser, and run it
        # logger.debug("Running parser %a (%a:%a)", item, self.file, self.seek)
        try:
            match item:
                case [*entries]:
                    for entry in entries:
                        out += self.get(entry)

                case str(s):
                    substr = self.__source[self.seek:][:len(s)]
                    # Always case fold when matching
                    if substr.lower() == s.lower():
                        self.seek += len(s)
                        out += [s]
                    else:
                        raise ParseError(f'Expected {item!r}, got {substr!r} (char {self.seek})')

                case None:
                    pass

                case ParseDirective():
                    out += item.run(self)

                case other:
                    if callable(item):
                        out += self.get(item())
                    else:
                        raise ValueError(f"Unexpected item: {other}")

        except ParseError as e:
            e.stk.append(item)
            raise e

        out = self.__merge_matches(out)

        # Parse directives can have a handler attached to them.
        # This handler is run on the matched result, and should return
        # a new result.
        if isinstance(item, ParseDirective) and (f := item.handler):
            # print('Running handler:', f, merged)
            result = f(out)
            # print("Result:", repr(result))
            out = result

        # if isinstance(merged, MatchLiteral):
        #     merged = merged.matched

        # if not isinstance(merged, list):
        #     return [merged]
        # else:
        return out

    def __merge_matches(self, entries: list[Any]) -> list[Any]:
        """
        Attempt to merge similar match nodes into larger nodes.

        For example, a list of matched strings could look like

        .. code-block:: python

            ['H', 'e', 'll', 'o']

        This would merge it into

        .. code-block:: python

            ['Hello']

        If any token which can't be merged appears in the list, simply
        return the original list.
        """
        fixed_entries = []
        s = ''
        for entry in entries:
            match entry:
                case str(c):
                    s += c
                case _:
                    if s != '':
                        fixed_entries.append(s)
                    fixed_entries.append(entry)
                    s = ''
        if s != '':
            fixed_entries.append(s)

        return fixed_entries

    def snapshot(self) -> ParserCombinatorSnapshot:
        """Create a snapshot of the parsers current state."""
        return ParserCombinatorSnapshot(
                seek=self.seek,
                creator=self)

    def restore(self, snapshot: ParserCombinatorSnapshot) -> None:
        """Restore a snapshot, altering the parsers state."""
        assert snapshot._creator == self
        self.seek = snapshot._seek

    def peek_string(self, max_len: int) -> str:
        """
        Return the upcomming string, for debugging purposes.

        :param max_len:
            Maximum length to return, may be shorter if there isn't
            enough characters left
        """
        return self.__source[self.seek:][:max_len]


T = TypeVar('T')


def const(x: T) -> T:
    """
    Return the value directly.

    Useful for handlers.
    """
    return x
