"""
Python wrapper around puppet strings.

This maps
`Puppet Strings <https://www.puppet.com/docs/puppet/7/puppet_strings.html>`_
JSON output onto python objects, using type-level reflection (see
:class:`Deserializable` for details).

Fields declared without defaults are required, those with defaults are
optional. Extra fields may be given, but will result in a warning
being logged (TODO make this configurable).

Puppet Strings output should follow
`this schema <https://github.com/puppetlabs/puppet-strings/blob/main/JSON.md>`_,
unfortunately, that document is out of date [#f1]_, so the following
is mostly based on observation.

Some of the fields defined here have an Optional type, all these can
be changed to simply having a default value, if that better reflects
the possibilities of the field. The same is true with the above types
swaped.

.. rubric:: Footnotes

.. [#f1] It's out of date as of writing this (2023-07-03)
"""

import subprocess
from typing import (
    Any,
    Optional,
    Protocol,
)
from dataclasses import dataclass, field
import logging
from .internal import Deserializable
import re
# from muppet.util import ILoggerAdapter


logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class DocStringTag(Deserializable):
    """
    A generic DocString tag.

    note, api, summary

    :param tag_name:
        Type of the tag, like 'param', 'api', ...

    :param text:
        Freeform text content of the tag.
    """

    tag_name:   str
    text:       str = ''

    types:      Optional[list[str]] = None

    @staticmethod
    def _key() -> str:
        raise NotImplementedError()


@dataclass(kw_only=True)
class DocStringParamTag(DocStringTag):
    """Tag with tag_name set to 'param'."""

    types: list[str]
    """
    If tag_name is 'param', then this will be a singleton list of the
    type (as a string). Untyped fields will be typed as 'Any'.
    """

    name: str
    """Parameter name."""

    @staticmethod
    def _key() -> str:
        return 'param'


@dataclass(kw_only=True)
class DocStringExampleTag(DocStringTag):
    """Tag with tag_name set to 'example'."""

    name: str
    """
    Most likely name of language of ``text``.

    Will be the empty string if not set.
    """

    @staticmethod
    def _key() -> str:
        return 'example'


@dataclass(kw_only=True)
class DocStringOverloadTag(DocStringTag):
    """
    Tag with tag_name set to 'overload'.

    These three will be set for puppet 4.x functions with overloads
    when tag_name is 'overload'.
    """

    name: str
    signature: str
    docstring: Optional['DocString'] = None
    defaults: dict[Any, Any] = field(default_factory=dict)

    @staticmethod
    def _key() -> str:
        return 'overload'


@dataclass(kw_only=True)
class DocStringOptionTag(DocStringTag):
    """Tag with tag_name set to 'option'."""

    opt_name:   str
    opt_text:   str
    opt_types:  str
    parent:     str
    name:       str

    @staticmethod
    def _key() -> str:
        return 'option'


@dataclass(kw_only=True)
class DocStringSeeTag(DocStringTag):
    """Tag with tag_name set to 'see'."""

    name:       str

    @staticmethod
    def _key() -> str:
        return 'see'


@dataclass(kw_only=True)
class DocString(Deserializable):
    """Documentation entry for something."""

    text: str
    tags: list[DocStringTag] = field(default_factory=list)

    @staticmethod
    def handle_tags(items: list[dict[str, Any]]) -> list[DocStringTag]:
        """
        Parse list of tag dictionaries.

        The type of a tag dictionary depends on the value of
        ``tag_name``.
        """
        result: list[DocStringTag] = []
        for object in items:
            # cls: type[DocStringTag]
            cls: type
            for c in DocStringTag.__subclasses__():
                if object['tag_name'] == c._key():
                    cls = c
                    break
            else:
                cls = DocStringTag
            try:
                result.append(cls(**object))
            except TypeError as e:
                logger.error("Bad tag set for tag object (class=%s) %s",
                             cls.__name__, object,
                             exc_info=e)
                raise e
        return result


@dataclass(kw_only=True)
class PuppetClass(Deserializable):
    """Documentation for a puppet class."""

    name:       str
    file:       str
    line:       int
    inherits:   Optional[str] = None
    docstring:  DocString
    defaults:   dict[str, str] = field(default_factory=dict)
    source:     str


@dataclass(kw_only=True)
class DataType(Deserializable):
    """Documentation for a data type."""

    name:       str
    file:       str
    line:       int
    docstring:  DocString
    defaults:   dict[str, str]
    # source: str


@dataclass(kw_only=True)
class DataTypeAlias(Deserializable):
    """Documentation for a type alias."""

    name:       str
    file:       str
    line:       int
    docstring:  DocString
    alias_of:   str
    # source: str


@dataclass(kw_only=True)
class DefinedType(Deserializable):
    """Documentation for a defined type."""

    name: str
    file: str
    line: int
    docstring: DocString
    defaults: dict[str, str] = field(default_factory=dict)
    source: str


@dataclass(kw_only=True)
class Provider(Deserializable):
    """Documentation for a resource type provider."""

    name:       str
    type_name:  str
    file:       str
    line:       int
    docstring:  DocString
    confines:   dict[str, str] = field(default_factory=dict)
    features:   list[str] = field(default_factory=list)
    defaults:   list[Any] = field(default_factory=list)
    commands:   dict[str, str] = field(default_factory=dict)


@dataclass(kw_only=True)
class ResourceTypeProperty(Deserializable):
    """Documentation for a property of a resource type."""

    name:         str
    description:  str
    values:       Optional[list[str]] = None
    aliases:      dict[str, str] = field(default_factory=dict)
    isnamevar:    bool = False
    default:      Optional[str] = None

    required_features: Optional[str] = None
    data_type: Optional[str] = None


@dataclass(kw_only=True)
class ResourceTypeParameter(Deserializable):
    """Documentation for a parameter of a resource type."""

    name:         str
    description:  Optional[str] = None
    values:       Optional[list[str]] = None
    aliases:      Any = field(default_factory=list)
    isnamevar:    bool = False
    default:      Optional[str] = None

    required_features: Optional[str] = None
    data_type: Optional[str] = None


@dataclass(kw_only=True)
class ResourceTypeFeature(Deserializable):
    """Documentation for a published feature of a resource type."""

    name:         str
    description:  str


@dataclass(kw_only=True)
class ResourceType(Deserializable):
    """Documentation for a resource type."""

    name:        str
    file:        str
    line:        int
    docstring:   DocString
    properties:  list[ResourceTypeProperty] = field(default_factory=list)
    parameters:  list[ResourceTypeParameter]
    features:    Optional[list[ResourceTypeFeature]] = None
    providers:   Optional[list[Provider]] = None


@dataclass(kw_only=True)
class Signature(Deserializable):
    """Documentation for a function signature."""

    signature: str
    docstring: DocString


@dataclass(kw_only=True)
class Function(Deserializable):
    """Documentation for a function."""

    name:        str
    file:        str
    line:        int
    type:        str  # Probably one of 'ruby3x', 'ruby4x', 'puppet'
    signatures:  list[Signature]
    docstring:   DocString
    defaults:    dict[str, str] = field(default_factory=dict)
    source:      str


@dataclass(kw_only=True)
class Task(Deserializable):
    """Documentation for a task."""

    name:           str
    file:           str
    line:           int
    docstring:      Optional[DocString] = None
    source:         str
    supports_noop:  bool
    input_method:   Any


@dataclass(kw_only=True)
class Plan(Deserializable):
    """Documentation for a plan."""

    name:       str
    file:       str
    line:       int
    docstring:  DocString
    defaults:   dict[str, Any]
    source:     str


@dataclass(kw_only=True)
class PuppetStrings(Deserializable):
    """Complete documentation for a Puppet module."""

    puppet_classes:     list[PuppetClass]
    data_types:         list[DataType]
    data_type_aliases:  list[DataTypeAlias]
    defined_types:      list[DefinedType]
    resource_types:     list[ResourceType]
    providers:          list[Provider]
    puppet_functions:   list[Function]
    puppet_tasks:       list[Task]
    puppet_plans:       list[Plan]


# --------------------------------------------------


def puppet_strings(path: str) -> bytes:
    """
    Run ``puppet strings`` on puppet module at path.

    Returns a bytes object rather than a :class:`PuppetStrings`
    object, to efficeiently writing the output to a cache.

    .. code-block:: python
       :caption: Example Invocation

        >>> PuppetStrings.from_json(puppet_strings("/etc/puppetlabs/code/modules/stdlib"))
    """
    # All this extra weird stuff with tempfiles and pipes since puppet
    # strings output errors on stdout, and only if the --out flag
    # isn't used.
    import tempfile

    tmpfile = tempfile.NamedTemporaryFile()
    logger = logging.getLogger('muppet.puppet-strings')
    # logger = ILoggerAdapter(logging.getLogger('muppet.puppet-strings'),
    #                         extra={'file': path})

    cmd = subprocess.Popen(
            ['puppet', 'strings', 'generate',
             '--format', 'json',
             '--out', tmpfile.name],
            cwd=path,
            stdout=subprocess.PIPE,
            text=True,
            )

    if not cmd.stdout:
        # TODO better error?
        raise Exception(cmd.returncode)

    for line in cmd.stdout:
        line = line.strip()
        # These debug levels are by best effort, and based on the
        # enum found here:
        # https://github.com/puppetlabs/puppet-strings/blob/afe75151f8b47ce33433c488e22ca508aa48ac7c/spec/unit/puppet-strings/yard/handlers/ruby/rsapi_handler_spec.rb#L105
        # Expected formatting from observing the output.
        # TODO document this special logger
        if m := re.match(r'^\[(\w+)]: (.*)', line):
            match m[1]:
                case "debug":  logger.debug(m[2])
                case "warn":   logger.warning(m[2])
                case "error":  logger.error(m[2])
                case _:        logger.warning(line)
        else:
            logger.info(line)

    return tmpfile.read()


class HasDocstring(Protocol):
    """Something which has a docstring attribute."""

    docstring: DocString


def isprivate(entry: HasDocstring) -> bool:
    """
    Is the given puppet declaration marked private.

    Assumes input is a dictionary as returned by puppet strings, one
    of the entries in (for example) 'puppet_classes'.

    Currently only checks for an "@api private" tag.
    """
    for tag in entry.docstring.tags:
        if tag.tag_name == 'api' and \
           tag.text == 'private':
            return True
    return False
